use std::fs::File;
use std::io::stderr;
use std::path::PathBuf;

use clap::{Parser, Subcommand};
use color_eyre::eyre::Context;
use color_eyre::Result;
use tracing::metadata::LevelFilter;
use tracing_error::ErrorLayer;
use tracing_subscriber::fmt;
use tracing_subscriber::prelude::*;

use kak_highlight::{client, daemon};

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
struct Cli {
    /// Set the level of verbosity
    #[arg(short, long, action = clap::ArgAction::Count)]
    verbose: u8,
    /// File to write log events to
    #[arg(long, value_name = "FILENAME")]
    log: Option<PathBuf>,
    /// Config file. Defaults to `$XDG_CONFIG_HOME/kak-highlight.toml`.
    #[arg(short, long, value_name = "FILENAME")]
    config: Option<PathBuf>,
    /// The command to run
    #[command(subcommand)]
    command: Command,
}

#[derive(Subcommand)]
enum Command {
    /// Send a request payload to a running daemon and wait for the response
    Request,
    /// Start a daemon
    Daemon {
        /// The Kakoune session this daemon belongs to
        session: String,
    },
}

#[tracing::instrument]
fn main() -> Result<()> {
    let cli = Cli::parse();

    {
        let filter_layer = match cli.verbose {
            0 => LevelFilter::WARN,
            1 => LevelFilter::INFO,
            2 => LevelFilter::DEBUG,
            _ => LevelFilter::TRACE,
        };

        let output_layer = if let Some(path) = cli.log {
            let f = File::create(&path)
                .wrap_err_with(|| format!("Failed to create log file '{}'", path.display()))?;
            fmt::layer().compact().with_writer(f).boxed()
        } else if cfg!(debug_assertions) {
            fmt::layer().pretty().with_writer(stderr).boxed()
        } else {
            fmt::layer().compact().with_writer(stderr).boxed()
        };

        tracing_subscriber::registry()
            .with(filter_layer)
            .with(ErrorLayer::new(fmt::format::Pretty::default()))
            .with(output_layer)
            .init();
    }

    let runtime_dir = std::env::var("XDG_RUNTIME_DIR")
        .map(PathBuf::from)
        .unwrap_or_else(|_| "/var".into())
        .join("kak-highlight");

    let res = match cli.command {
        Command::Daemon { session } => daemon::handle(runtime_dir, cli.config, session),
        Command::Request => client::handle(runtime_dir),
    };

    if let Err(err) = &res {
        tracing::error!("{:?}", err)
    }

    res
}
